@@ -1,12 +1,12 @@
 import argparse
 from tqdm import tqdm
 from sklearn.model_selection import train_test_split
-from nlp import load_dataset
+from datasets import load_dataset
 import urllib.request
 import os, json, sys
 import shutil
 import tarfile
-
+from .utils import add_filehandler_for_logger
 # these have been tested and work fine. more can be added to this list to test
 MODELS_SUPPORT = ["distilbert-base-cased", "bert-base-uncased", "bert-large-uncased",
                   "google/electra-base-discriminator",
@@ -117,8 +117,26 @@ def data_check(args, sample_val_from_train=True, val_sample_portion=0.1,logger=N
                 print("data not found")
                 sys.exit(0)
 
+def check_output_path(output_path,force=False):
+    if os.path.isdir(output_path):
+        if force:
+            print(f"{output_path} exists, remove it as force=True")
+            shutil.rmtree(output_path)
+            os.makedirs(output_path, exist_ok=True)
+        else:
+            out = input(
+                "Output directory ({}) already exists and is not empty, you wanna remove it before start training? (y/n)".format(
+                    output_path))
+            if out.lower() == "y":
+                shutil.rmtree(output_path)
+                os.makedirs(output_path, exist_ok=True)
+            else:
+                sys.exit(0)
+    else:
+        print(f"{output_path} not found, create it now")
+        os.makedirs(output_path, exist_ok=True)
 
-def sanity_check(args,logger=None):
+def sanity_check(args,logger=None,force=False):
     # auto-construct some args
     # check if data exists
     data_check(args,logger=logger)
@@ -129,17 +147,7 @@ def sanity_check(args,logger=None):
     args.output_path = output_path
 
     if args.do_train:
-        if os.path.isdir(output_path):
-            out = input(
-                "Output directory ({}) already exists and is not empty, you wanna remove it before start training? (y/n)".format(
-                    output_path))
-            if out.lower() == "y":
-                shutil.rmtree(output_path)
-                os.makedirs(output_path, exist_ok=True)
-            else:
-                sys.exit(0)
-        else:
-            os.makedirs(output_path, exist_ok=True)
+        check_output_path(output_path,force=force)
 
     assert args.model_select in MODELS_SUPPORT or os.path.isdir(args.model_select), F"set --model_select has to be in {MODELS_SUPPORT} or a local path where model's config and tokenizer_config exist"
 
@@ -156,7 +164,8 @@ def sanity_check(args,logger=None):
 
     if "pretrain" in args.task:
         assert "t5" in args.model_select, "pretrain task now is only compatible with T5 models so far"
-
+    if logger is not None:
+        add_filehandler_for_logger(args.output_path, logger)
 
 class Args:
     '''
@@ -164,6 +173,7 @@ class Args:
     '''
     model_select = "bert-base-uncased"
     data_path = "data/glue/sst2"
+    dataset_name = ","
     task = "single-label-cls"
     per_device_train_batch_size = 8
     eval_batch_size = 32
@@ -204,6 +214,9 @@ def get_args():
     parser.add_argument('--data_path', type=str, default="data/glue/sst2",
                         help='data path')
 
+    parser.add_argument('--dataset_name', type=str, default="",
+                        help="dataset name for HF's load_dataset")
+
     parser.add_argument('--task', type=str, default="t2t",
                         help='tasks available in TASKS_SUPPORT')
 
